package gcewing.codechicken.lib.render;

import java.nio.IntBuffer;
import java.util.ArrayList;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import gcewing.codechicken.lib.render.SpriteSheetManager.SpriteSheet;
import gcewing.codechicken.lib.render.TextureUtils.ITUIconRegister;
import cpw.mods.fml.relauncher.Side;
import cpw.mods.fml.relauncher.SideOnly;
import net.minecraft.client.Minecraft;
import net.minecraft.client.renderer.GLAllocation;
import net.minecraft.client.renderer.texture.IIconRegister;
import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.texture.TextureMap;
import net.minecraft.client.renderer.texture.TextureUtil;
import net.minecraft.client.resources.IResourceManager;
import net.minecraft.util.ResourceLocation;

@SideOnly(Side.CLIENT)
public class TextureSpecial extends TextureAtlasSprite implements ITUIconRegister
{
    //sprite sheet fields
    private int spriteIndex;
    private SpriteSheet spriteSheet;
    
    //textureFX fields
    private TextureFX textureFX;
    
    private int blankSize = -1;

    private ArrayList<TextureDataHolder> baseTextures;
    
    private boolean selfRegister;
    public int atlasIndex;
    
    protected TextureSpecial(String par1)
    {
        super(par1);
    }
    
    public TextureSpecial addTexture(TextureDataHolder t)
    {
        if(baseTextures == null)
            baseTextures = new ArrayList<TextureDataHolder>();
        baseTextures.add(t);
        return this;
    }
    
    public TextureSpecial baseFromSheet(SpriteSheet spriteSheet, int spriteIndex)
    {
        this.spriteSheet = spriteSheet;
        this.spriteIndex = spriteIndex;
        return this;
    }
    
    public TextureSpecial addTextureFX(TextureFX fx)
    {
        textureFX = fx;
        return this;
    }
    
    @Override
    public void initSprite(int sheetWidth, int sheetHeight, int originX, int originY, boolean rotated)
    {
        super.initSprite(sheetWidth, sheetHeight, originX, originY, rotated);
        if(textureFX != null)
            textureFX.onTextureDimensionsUpdate(width, height);
    }
    
    @Override
    public void updateAnimation()
    {
        if(textureFX != null)
        {
            textureFX.update();
            if(textureFX.changed())
                uploadTextureSub(textureFX.imageData, width, height, originX, originY, false, false, false);
        }
    }
    
    @Override
    public boolean load(IResourceManager manager, ResourceLocation location)
    {
        if(baseTextures != null)
        {
            for(TextureDataHolder tex : baseTextures)
            {
                framesTextureData.add(tex.data);
                width = tex.width;
                height = tex.height;
            }
        }
        
        if(spriteSheet != null)
        {
            TextureDataHolder tex = spriteSheet.createSprite(spriteIndex);
            width = tex.width;
            height = tex.height;
            framesTextureData.add(tex.data);
        }
        
        if(blankSize > 0)
        {
            width = height = blankSize;
            framesTextureData.add(new int[blankSize*blankSize]);
        }
        
        if(framesTextureData.isEmpty())
            throw new RuntimeException("No base frame for texture: "+getIconName());
        
        return true;
    }
    
    @Override
    public boolean hasAnimationMetadata()
    {
        return textureFX != null || super.hasAnimationMetadata();
    }
    
    @Override
    public int getFrameCount()
    {
        if(textureFX != null)
            return 1;
        
        return super.getFrameCount();
    }

    public TextureSpecial blank(int size)
    {
        blankSize = size;
        return this;
    }
    
    public TextureSpecial selfRegister()
    {
        selfRegister = true;
        TextureUtils.addIconRegistrar(this);
        return this;
    }
    
    @Override
    public void registerIcons(IIconRegister register)
    {
        if(selfRegister)
            ((TextureMap)register).setTextureEntry(getIconName(), this);
    }
    
    @Override
    public int atlasIndex()
    {
        return atlasIndex;
    }

    private static final IntBuffer dataBuffer = GLAllocation.createDirectIntBuffer(4194304);

    private static void uploadTextureSub(int[] imageData, int width, int height, int originX, int originY, boolean linear, boolean clamped, boolean mipmapped)
    {
        int j1 = 4194304 / width;
        setTextureInterpolation(linear, mipmapped);
        setTextureClamped(clamped);
        int i2;

        for (int k1 = 0; k1 < width * height; k1 += width * i2)
        {
            int l1 = k1 / width;
            i2 = Math.min(j1, height - l1);
            int j2 = width * i2;
            copyToBufferPos(imageData, k1, j2);
            GL11.glTexSubImage2D(GL11.GL_TEXTURE_2D, 0, originX, originY + l1, width, i2, GL12.GL_BGRA, GL12.GL_UNSIGNED_INT_8_8_8_8_REV, dataBuffer);
        }
    }

    private static void setTextureInterpolation(boolean linear, boolean mipmapped)
    {
        if (linear)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, mipmapped ? GL11.GL_LINEAR_MIPMAP_LINEAR : GL11.GL_LINEAR);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        }
        else
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, mipmapped ? GL11.GL_NEAREST_MIPMAP_LINEAR : GL11.GL_NEAREST);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
        }
    }

    private static void setTextureClamped(boolean clamped)
    {
        if (clamped)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_CLAMP);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_CLAMP);
        }
        else
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_REPEAT);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_REPEAT);
        }
    }

    private static void copyToBufferPos(int[] data, int par1, int par2)
    {
        if (Minecraft.getMinecraft().gameSettings.anaglyph)
        {
            data = TextureUtil.updateAnaglyph(data);
        }
        dataBuffer.clear();
        dataBuffer.put(data, par1, par2);
        dataBuffer.position(0).limit(par2);
    }

}
